def factorials(n: int):
    k = 1
    facts = [1]
    for j in range(1, n + 1):
        k *= j
        facts.append(k)
    return facts

def n_over_ks(n: int):
    facts = factorials(n)
    return [[facts[j]//(facts[k]*facts[j - k]) for k in range(j + 1)] for j in range(len(facts))]

def sum_by_name(ls, name='jetpolynom'):
    # sum over the members of a given list so that each member e having e.__class__.__name__ appears on the
    # left. This can be used to prevent that in case of numpy the new values are numpy.arrays.
    return sum([e for e in ls if e.__class__.__name__ == name]) + sum([e for e in ls if e.__class__.__name__ != name])

def general_leibnitz_rule(f1, f2):
    '''
    Compute the higher-order derivatives of the product of two functions f1*f2.
    In the following denote by f^k the k-th derivative of f, evaluated at a specific point.
    It is assumed that len(f1) = len(f2) =: n.
    
    Parameters
    ----------
    f1: list
        List containing the values f1^k for k = 0, ..., n - 1.
    f2: list
        List containing the values f2^k for k = 0, ..., n - 1.
        
    Returns
    -------
    list
        List containing the values (f1*f2)^k for k = 0, ..., n - 1.
    '''
    nmax = len(f1) - 1 # len(f1): max number of summands
    nok = n_over_ks(nmax)
    return [sum_by_name([nok[n][k]*f1[n-k]*f2[k] if f1[n-k].__class__.__name__ == 'jetpolynom' else nok[n][k]*f2[k]*f1[n-k] for k in range(n + 1)]) for n in range(nmax + 1)]

def faa_di_bruno(f, g):
    '''
    Consider the composition f o g of two functions and for h in [f, g] denote by h^k its k-th derivative. Then
    (f o g)^1 = (f^1 o g) g^1
    (f o g)^2 = (f^2 o g) (g^1)**2 + (f^1 o g) g^2
    (f o g)^3 = ... and so on.
    
    Let F = [f o g, (f^1 o g), (f^2 o g), ...] and G = [g, g^1, g^2, ...] be given. Then this
    function will compute (f o g)^k up to the given (common) order, i.e. the
    chain rule for the k-th derivative.
    
    This routine hereby will make use of the equation
    (f^n o g) = sum_(k = 1)^n f^k B(n, k)(g^1, g^2, ..., g^(n - k + 1)) ,  
    where B(n, k) for n >= k are the incomplete exponential Bell polynomials.
    
    Parameters
    ----------
    f: list
        List containing the values F (see above).
    g: list
        List containing the values G (see above).
        
    Returns
    -------
    dfdg: list
        List containing the values (f o g)^k for k = 0, ..., len(f) - 1.
    '''
    bell = bell_polynomials(len(f) - 1, g[1:])
    return [sum_by_name([bell.get((n, k), 0)*f[k] for k in range(len(f))]) for n in range(len(f))] # or with numpy arrays: np.matmul(bell, f)

def bell_polynomials(n: int, z):
    '''
    Compute the incomplete exponential Bell polynomials B(n, k, [z1, ..., z(n - k + 1)]) for all 1 <= k <= n.

    Parameters
    ----------
    n: int
        The maximal order to which the Bell polynomials should be computed.
    z: subscriptable
        List of values at which the Bell polynomials are supposed to be evaluated. Note that it must hold: len(z) == n.

    Returns
    -------
    bp: dict
        dict of Bell polynomials up to the given order n. Its keys denote the indices (n, k) of the polynomials.
    '''
    assert len(z) == n
    B = {}
    B[(0, 0)] = 1.0
    nok = n_over_ks(n)
    for jn in range(n + 1):
        for jk in range(1, jn + 1):
            B[(jn, jk)] = sum_by_name([nok[jn - 1][m - 1]*B.get((jn - m, jk - 1), 0)*z[m - 1] for m in range(1, jn - jk + 2)])
    return B

    
class jet:
    def __init__(self, value=0, **kwargs):        
        if hasattr(value, '__getitem__'):
            self.set_array(value, **kwargs)
        else:
            self.set_array([value], **kwargs)
        self.graph = kwargs.get('graph', [(1, '0'), self]) # for keeping track of the computation graph
        
    def get_array(self, **kwargs):
        n = kwargs.get('n', self.order)
        return [self.array(k) for k in range(n + 1)]
    
    def set_array(self, array, **kwargs):
        if hasattr(array, '__len__'):
            omax = len(array) - 1
            self.array = lambda k: array[k] if k <= omax else 0
        else:
            omax = 0
            self.array = lambda k: array[k]
        self.set_order(n=kwargs.get('n', omax))        
        
    def set_order(self, n):
        self.order = n
        
    def __str__(self):
        outstr = ''
        for e in self.get_array():
            if not isinstance(e, type(self)):
                outstr += e.__str__()
            else:
                outstr += str(e)
            outstr += ', '
        return f'{self.order}-jet({outstr[:-2]})'

    def _repr_html_(self):
        return f'<samp>{self.__str__()}</samp>'
    
    def __neg__(self):
        result = jet(n=self.order, graph=[(1, '-'), self.graph])
        result.array = lambda n: -self.array(n)
        return result
    
    def __add__(self, other):
        other = convert(other)
        max_order = max([self.order, other.order])
        result = jet(n=max_order, graph=[(2, '+'), self.graph, other.graph])
        result.array = lambda n: self.array(n) + other.array(n)
        return result
    
    def __radd__(self, other):
        other = convert(other)
        return other + self
    
    def __sub__(self, other):
        return self + -other
    
    def __rsub__(self, other):
        return other + -self
        
    def __mul__(self, other):
        other = convert(other)
        max_order = max([self.order, other.order])
        result = jet(n=max_order, graph=[(2, '*'), self.graph, other.graph])
        # compute the derivatives      
        f1, f2 = self.get_array(n=max_order), other.get_array(n=max_order)
        glr = general_leibnitz_rule(f1, f2)
        result.array = lambda n: glr[n] if n <= max_order else 0
        # result.array = lambda k: general_leibnitz_rule(f1[:k + 1], f2[:k + 1]) if k <= max_order else 0
        # The next line would work for arbitrary orders, but it is also much slower instead of pre-loading the array:
        # result.array = lambda n: sum([n_over_k(n, k)*self.array(n - k)*other.array(k) for k in range(n + 1)])
        return result
    
    def __rmul__(self, other):
        other = convert(other)
        return other*self
    
    def __pow__(self, other):
        '''
        N.B.: Power defined here only for exponents in the given field (integer, float, complex numbers),
        not jets themselves as exponents.
        '''
        other = convert(other)
        f = self.array(0)
        g = other.array(0)
        n = self.order
        
        # compute the derivatives
        n_additional_zeros = 0
        for nmax in range(n + 1):
            if g - nmax == 0:
                n_additional_zeros = n - nmax
                break
        # nmax defines the depth of the iteration. Any remaining derivatives are set to zero.

        # fundamental equation:
        #  fp[(y, n)] = y*sum([n_over_k(n - 1, k)*fp[(y - 1, n - 1 - k)]*f[k + 1] for k in range(n)]),
        # so fp[(y, n)] depends on fp[(y - 1, n - 1)], fp[(y - 1, n - 2)], ..., fp[(y - 1, 0)]. This means for each "layer" with respect to y
        # we need to compute the entire set of derivatives up to the (n-1)th order. The iteration start begins with n=1 at the lowest layer.
        # The individual layers are labelled by the integer max_der.
        fp = {}
        fp[(0, 0)] = f**(g - nmax)
        layer = g - nmax + 1
        nok = n_over_ks(nmax - 1)
        for max_der in range(1, nmax + 1):
            fp[(max_der, 0)] = f**layer
            for order in range(1, max_der + 1):           
                fp[(max_der, order)] = layer*sum_by_name([nok[order - 1][k]*self.array(k + 1)*fp[(max_der - 1, order - 1 - k)] for k in range(order)])
            layer += 1
        # extract the derivatives
        df = [fp[(nmax, order)] for order in range(0, nmax + 1)] + [0]*n_additional_zeros
        result = jet(df)
        result.graph = [(2, '**'), self.graph, other.graph]
        return result
    
    def __rpow__(self, other):
        other = convert(other)    
        return other**self
    
    def __truediv__(self, other):
        other = convert(other)
        return self.__mul__(other.inv())
    
    def __rtruediv__(self, other):
        other = convert(other)
        return other.__mul__(self.inv())
    
    def copy(self):
        '''
        Return a copy of this jet.
        '''
        result = jet(n=self.order, graph=self.graph)
        result.array = self.array
        return result
    
    def derive(self):
        '''
        Return the derivative of this jet, determined by shifting its array to the left by 1.
        '''
        array = self.get_array()
        result = self.copy()
        if len(array) == 1:
            result.set_array([0])
        else:
            result.set_array(array[1:])
        return result
    
    def compose(self, other):
        '''
        Compute the composition of two jets, based on Faà di Bruno's formula.
        It is assumed that the array ar with ar = self.get_array() belongs to the values
        [f o g, f^1 o g, f^2 o g, ... ], where f^k denote the k-th derivative of the function f,
        so this function will compute the values
        [f o g, (f o g)^1, (f o g)^2, ...]
        '''        
        f = self.get_array()
        g = other.get_array()
        dfg = faa_di_bruno(f, g)
        result = self.copy()
        result.set_array(dfg)
        return result
    
    def inv(self):
        '''
        Compute 1/jet. This will be done by applying the chain rule to
        "1/x o jet".
        '''
        f = self.array(0)
        assert f != 0

        facts = factorials(self.order)
        invf = [(-1)**n*facts[n]/f**(n + 1) for n in range(self.order + 1)]
        result = self.copy()
        result.set_array(invf)
        result = result.compose(self)
        result.graph = [(1, '1/'), self.graph]
        return result

    
def convert(other):
    '''
    Convert an object to a (constant) jet.
    '''
    if not other.__class__.__name__ == 'jet':
        return jet(value=other)
    else:
        return other
        
    



# Collection of elementary functions whose derivatives are known to any order.

from .jet import jet

import numpy as np
import sympy

def detect_code(x):
    # Find the appropriate code to perform operations.
    # We shall assume that the code labels are identical (otherwise we may need to
    # implement a dictionary translating between the various names).
    code = np
    if hasattr(x, '__module__'):
        if x.__module__[0:5] == 'sympy':
            code = sympy
    return code

def sin(x, **kwargs):
    '''
    Compute the sin of a jet.

    Parameters
    ----------
    x: jet

    Returns
    -------
    jet
    '''
    x0 = x.array(0)
    code = kwargs.get('code', detect_code(x0))
    s = code.sin(x0)
    c = code.cos(x0)
    result = jet(s, n=x.order, graph=[(1, 'sin'), x.graph])

    # compute the derivatives
    result.array = lambda n: [s, c, -s, -c][n%4]
    return result.compose(x)

def cos(x, **kwargs):
    '''
    Compute the cos of a jet.

    Parameters
    ----------
    x: jet

    Returns
    -------
    jet
    '''
    x0 = x.array(0)
    code = kwargs.get('code', detect_code(x0))
    s = code.sin(x0)
    c = code.cos(x0)
    result = jet(c, n=x.order, graph=[(1, 'cos'), x.graph])

    # compute the derivatives
    result.array = lambda n: [c, -s, -c, s][n%4]
    return result.compose(x)

def exp(x, **kwargs):
    '''
    Compute the exponential of a jet.

    Parameters
    ----------
    x: jet

    Returns
    -------
    jet
    '''
    x0 = x.array(0)
    code = kwargs.get('code', detect_code(x0))
    e = code.exp(x0)
    result = jet(e, n=x.order, graph=[(1, 'exp'), x.graph])

    # compute the derivatives
    result.array = lambda n: e
    return result.compose(x)

def log(x, **kwargs):
    '''
    Compute the natural logarithm of a jet.

    Parameters
    ----------
    x: jet

    Returns
    -------
    jet
    '''
    x0 = x.array(0)
    code = kwargs.get('code', detect_code(x0))
    ln = code.log(x0)
    graph=[(1, 'log'), x.graph]

    # compute the derivatives
    dx = x.copy().derive()
    drx_arr = (dx/x).get_array()[:-1]
    result = jet(n=x.order, graph=graph)
    result.set_array([ln] + drx_arr)
    return result

def sinh(x, **kwargs):
    '''
    Compute the sinh of a jet.

    Parameters
    ----------
    x: jet

    Returns
    -------
    jet
    '''
    x0 = x.array(0)
    code = kwargs.get('code', detect_code(x0))
    sh = code.sinh(x0)
    ch = code.cosh(x0)
    result = jet(sh, n=x.order, graph=[(1, 'sinh'), x.graph])

    # compute the derivatives
    result.array = lambda n: [sh, ch][n%2]
    return result.compose(x)

def cosh(x, **kwargs):
    '''
    Compute the cosh of a jet.

    Parameters
    ----------
    x: jet

    Returns
    -------
    jet
    '''
    x0 = x.array(0)
    code = kwargs.get('code', detect_code(x0))
    ch = code.cosh(x0)
    sh = code.sinh(x0)
    result = jet(ch, n=x.order, graph=[(1, 'cosh'), x.graph])

    # compute the derivatives
    result.array = lambda n: [ch, sh][n%2]
    return result.compose(x)




def check_zero(value):
    # check if a value is zero; value may be an iterable
    if not hasattr(value, '__iter__'):
        return value == 0
    else:
        return all(value == 0)

class jetpolynom:
    '''
    Implementation of a polynom with arbitrary number of variables.
    
    The information of the coefficients is stored internally as dictionary in 
    the 'values' field.

    A polynomial p can be initiated with a value v and (optional) its index (describing
    the variable) and power n so that overall p = v*x_i**n. A general polynomial in several
    variables can then be constructed by successive addition and multiplication of different 
    classes with each other.
    
    Example
    -------    
    | self.values = {
    | frozenset({(0, 1)}): 11,
    | frozenset({(0, 4), (1, 22), (13, 1)}): 89,
    | frozenset({(4, 9), (7, 2), (12, 33)}): -0.22
    | }
    
    | The interpretation is:
    | 11*x0**1 +
    | 89*x0**4*x1**22*x13**1 +
    | -0.22*x4**9*x7**2*x12**33
    '''
    def __init__(self, value=0, index: int=0, power: int=0, **kwargs):
        self.values = kwargs.get('values', {frozenset([(index, power)]): value})
        
    def __str__(self):
        outstr = '['
        for key, value in self.values.items():
            fac = ''
            for e in key:
                if e[1] == 0: # do not show z**0
                    continue
                fac += f'*x{e[0]}**{e[1]}'
            outstr += f'{value}{fac} + \n'
        outstr = outstr[:-4]
        outstr += ']'
        return outstr
    
    def _repr_html_(self):
        outstr = self.__str__().replace('\n', '<br>')
        return f'<samp>{outstr}</samp>'
    
    def __add__(self, other):
        if not other.__class__.__name__ == 'jetpolynom':
            other = jetpolynom(other)
        new_values = {}
        for k in set(self.values).union(set(other.values)):
            new_value = self.values.get(k, 0) + other.values.get(k, 0)
            if check_zero(new_value): # do not store 0-values; x may be float or numpy array etc.
                continue
            new_values[k] = new_value
        if len(new_values) == 0:
            return jetpolynom(0)
        else:
            return jetpolynom(values=new_values)
    
    def __radd__(self, other):
        other = jetpolynom(other)
        return other + self
    
    def __neg__(self):
        new_values = {}
        for key, value in self.values.items():
            new_values[key] = -value
        return jetpolynom(values=new_values)
    
    def __sub__(self, other):
        return self + -other
    
    def __rsub__(self, other):
        return other + -self
    
    def __mul__(self, other):
        # Interpretation: (sum_j a_j)*(sum_k b_k) = sum_{j, k} a_j*b_k
        if not other.__class__.__name__ == 'jetpolynom':
            other = jetpolynom(other)
        pol_prod = {}
        for aj, value1 in self.values.items():
            e1 = dict(aj) # e.g. e1 = {0: 0, 1: 5, 2: 3}
            for bk, value2 in other.values.items():
                value_prod = value1*value2
                e2 = dict(bk) # e.g. e2 = {0: 0, 1: 0, 2: 1}
                e_prod = frozenset([(k, e1.get(k, 0) + e2.get(k, 0)) for k in set(e1).union(set(e2))]) # e.g. e_prod = frozenset([(0, 0), (1, 5), (2, 4)])
                value_prod += pol_prod.get(e_prod, 0) # account for multiplicity
                pol_prod[e_prod] = value_prod
        pol_prod = {k: v for k, v in pol_prod.items() if not check_zero(v)} # remove zero values
        return jetpolynom(values=pol_prod)
    
    def __rmul__(self, other):
        other = jetpolynom(other)
        return other*self
    
    def __pow__(self, other):
        assert (type(other) == int) and (other >= 0)
        if other == 0:
            return jetpolynom(1) # N.B. 0**0 := 1

        remainder = other%2
        half = self**(other//2)
        if remainder == 1:
            return self*half*half
        else:
            return half*half




from .functions import exp, log
from .jet import jet as jet_source
from .jet import factorials
from .poly import jetpolynom


class jet(jet_source):
    '''
    Class to store and operate with higher-order derivatives of a given function.
    '''
    def __pow__(self, other):
        if not isinstance(other, jet):
            other = jet(value=other) # n.b. convert from ad.py would convert to 'jet_source', not jet'.

        if other.order == 0:
            result = jet_source.__pow__(self, other)
        else:
            '''
            General exponentiation, using exp and log functions.
            '''
            result = exp(other*log(self))
            result.graph = [(2, '**'), self.graph, other.graph]
        return result
    
    def __rpow__(self, other):
        if not isinstance(other, jet):
            other = jet(value=other) # n.b. convert from ad.py would convert to 'jet_source', not jet'.
        return other**self
    
    
class derive:
    '''
    Class to handle the derivatives of a (jet-)function (i.e. a function consisting of a composition
    of elementary functions).
    '''
    
    def __init__(self, func, order: int=1, **kwargs):
        self.func = func
        self.n_args = self.func.__code__.co_argcount # the number of any arguments of func (before *args)
        self.set_order(order)
        self._Df = {}
        
    def set_order(self, order):
        self.order = order
        self._factorials = factorials(self.order)
        
    def eval(self, z, mult=True):
        '''
        Evaluate the derivatives of a (jet-)function at a specific point up to self.order.

        Parameters
        ----------
        z: subscriptable 
            Vector at which the function and its derivatives should be evaluated.
        mult: Boolean, optional
            Whether or not to include the factorials C(j1, ..., jm) default: True. (Details see below).
            If False, then the C*Df's are returned. If True, then the Df's are returned.

        Returns
        -------
        dict
            Dictionary of compontens of the multivariate Taylor expansion of the given function self.func:

            Let m be the number of arguments of self.func. 
            Then
            Df_k(z1, ..., zm) = sum_{j1 + ... + jm = k} C(j1, ..., jm) * Df[j1, ... jm] * z1**j1 * ... * zm**jm
            with
            Df[j1, ..., jm] := \partial^j1/\partial_{z1}^j1 ... \partial^jm/\partial_{zm}^jm f (z1, ..., zm)
            and combinatorial factor
            C(j1, ..., jm) = (j1 + ... + jm)!/(j1! * ... * jm!) .
        '''
        # perform the computation, based on the input vector
        inp = []
        for k in range(self.n_args):
            jk = jet([z[k], jetpolynom(1, index=k, power=1)], n=self.order)
            inp.append(jk)
        evaluation = self.func(*inp)
        
        # extract Df from the result
        Df = {}
        for k in range(1, self.order + 1): # the k-th derivative
            polynomials_k = evaluation.array(k).values
            for key, value in polynomials_k.items(): # loop over the individual polynomials of the k-th derivative
                # key corresponds to a specific frozenset, i.e. some indices and powers of a specific monomial.
                indices = [0]*self.n_args
                multiplicity = 1
                for tpl in key:
                    if tpl == (0, 0): # the (0, 0)-entries correspond to the scalar 1 and will be ignored here. TODO: may need to improve this in jetpolynom class.
                        continue
                    index, power = tpl
                    indices[index] = power
                    multiplicity *= self._factorials[power]
                if mult:
                    value *= multiplicity/self._factorials[sum(indices)]
                Df[tuple(indices)] = value
                
        self._Df = Df
        return Df
    
    def extract(self, k: int, **kwargs):
        '''
        Extract the components of the k-th derivative from the output of self.eval. 
        
        Parameters
        ----------
        k: int
            The order of the derivatives to extract.
        Df: dict, optional
            The output of self.eval, containing the entries of the derivative. If nothing
            specified, the most recent output will be used.
            
        Returns
        -------
        dict
            The components of the k-th derivative of self.func.
        '''
        assert k <= self.order
        D = kwargs.get('Df', self._Df)
        return {j: D[j] for j in D.keys() if sum(j) == k}
    
    @staticmethod
    def convert_indices(list_of_tuples):
        '''
        Convert a list of tuples denoting the indices in a multivariate Taylor expansion into a list of indices of
        the corresponding multilinear map.
        
        Parameters
        ----------
        list_of_tuples: list
            List of (self.n_args)-tuples denoting the indices in the multivariate Taylor expansion.
        
        Returns
        -------
        list
            List of tuples denoting the indices of a multilinear map.
        
        Example
        -------
        list_of_tuples = [(0, 2, 1, 9), (1, 0, 3, 0)]
        In this example we are looking at indices in the Taylor expansion of a function in 4 variables.
        The first member, (0, 2, 1, 9), corresponds to x0**0*x1**2*x2*x3**9 which belongs to the multilinear map
        of order 0 + 2 + 1 + 9 = 12, the second member to x0*x2**3, belonging to the multilinear map of order
        1 + 3 = 4.
        Hence, these indices will be transformed to tuples of length 12 and 4, respectively:
        (1, 1, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3) (variable 1 has power 2, variable 2 has power 1, variable 3 has power 9) 
        and
        (0, 2, 2, 2) (variable 0 has power 1, variable 2 has power 3).
        '''
        l1g = [[tpl[j]*[j] for j in range(len(tpl))] for tpl in list_of_tuples]
        return [tuple(e for subl in l1 for e in subl) for l1 in l1g] # flatten l1 and return list of converted tuples
    
    def build_tensor(self, k: int, **kwargs):
        '''
        Convert the components of the k-th derivative into the entries of a (self.n_args)**k tensor.
        See also self.convert_indices for details.
        
        Parameters
        ----------
        k: int
            The degree of the derivatives to be considered.
        
        Returns
        -------
        dict
            Dictionary representing non-zero components of the tensor which describes the multivariate 
            Taylor map of order k.
            Only values unequal to zero will be stored, and only one member for each set of entries
            which can be obtained by suitable permutations of the indices.
        '''
        entries = self.extract(k=k, **kwargs)
        list_of_tuples = list(entries.keys())
        converted_indices = self.convert_indices(list_of_tuples)
        return {converted_indices[k]: entries[list_of_tuples[k]] for k in range(len(list_of_tuples))}
    
    def grad(self, **kwargs):
        '''
        Returns the gradient of the function.
        
        Parameters
        ----------
        kwargs: dict
            Additional arguments passed to self.build_tensor
            
        Returns
        -------
        dict
            Dictionary containing the components of the gradient.
        '''
        return self.build_tensor(k=1, **kwargs)
    
    def hess(self, **kwargs):
        '''
        Returns the Hessian of the function.
        
        Parameters
        ----------
        kwargs: dict
            Additional arguments passed to self.build_tensor
            
        Returns
        -------
        dict
            Dictionary containing the components of the Hessian.
        '''
        return self.build_tensor(k=2, **kwargs)
        
        
        
